// Copyright (c) 2016, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by a MIT-style license
// that can be found in the LICENSE file.

part of omnibuild;

/// Message property information
class OmnibuildProperty {
  /// Property name
  final String name;

  /// Property docomentation
  final String documentation;

  /// Property type name
  final String type;

  /// Primitive type this property occupies in a JSON data map.
  final String jsonType;

  /// Indicate if the property is an enumeration.
  final bool enumeration;

  /// Indicate if the property is passed as named argument.
  final bool named;

  /// Indicate the default value of the property.
  final String defaultValue;

  /// Constructor
  OmnibuildProperty(this.name, this.documentation, this.type, this.jsonType,
      this.enumeration, this.named, this.defaultValue);

  /// Check if the given type should be validated by _getJsonValue.
  bool isJsonType() {
    return enumeration ? true : _isJsonType(type);
  }

  /// Commented documentation
  ///
  /// Add two leading spaces because properties are class members and all class
  /// members are incremented by two spaces (including the documentation).
  String get commentedDocumentation =>
      _prefixEachLine(documentation, '  /// ', false);

  /// Get the Dart code that should be used to reconstruct this type from the
  /// data map (contains primitive data only)
  ///
  /// This getter is used in the code template and it is assumed that you can
  /// retrieve the value using `data['propertyName']`.
  /// Note that impure Maps are not supported and should be filtered out in the
  /// [_parsePropertyYaml] function.
  String reverseConstruction() {
    if (enumeration) {
      return "$type.values[data['$name'] as int]";
    } else {
      return _createReverseConstructor(type, "data['$name']", jsonType, 1);
    }
  }
}

/// Message information
class OmnibuildMessage {
  /// Full omnibusId
  final String omnibusId;

  /// Message name
  final String name;

  /// Message documentation
  final String documentation;

  /// Message type (event, request, response)
  final OmnibusMessageType type;

  /// Message properties
  final List<OmnibuildProperty> properties;

  /// Constructor
  OmnibuildMessage(this.omnibusId, this.name, this.documentation, this.type,
      this.properties);

  /// Check if there is a property that is not pure JSON data
  ///
  /// If all types are pure JSON data this message can be initialized without
  /// type checking by [getJsonValue] (private function in Omnibus).
  bool supportsUncheckedInit() {
    return properties.where((OmnibuildProperty p) => !p.isJsonType()).isEmpty;
  }

  /// Generate Dart code for the OmnibusMessage constructor arguments.
  String argumentsString() {
    var positionalParameters = [];
    var optPositionalParameters = [];
    var namedParameters = [];

    // Code for passing the parameters to the key-value message map
    var mapValues = [];

    // Read properties.
    properties.forEach((OmnibuildProperty property) {
      /// Add property to a parameter group.
      if (property.named) {
        // Add parameter to named parameters.
        namedParameters.add(
            'final ${property.type} ${property.name}: ${property.defaultValue}');
      } else {
        // Add parameter to optional positional parameters if a default value is
        // defined.
        if (property.defaultValue.isEmpty) {
          positionalParameters.add('final ${property.type} ${property.name}');
        } else {
          optPositionalParameters.add(
              'final ${property.type} ${property.name} = ${property.defaultValue}');
        }
      }

      // Add parameter to map values.
      mapValues.add(
          "'${property.name}': ${property.enumeration ? '${property.name}.index' : property.name}");
    });

    // Generate parameters string.
    var parameters = positionalParameters.join(',');
    if (optPositionalParameters.isNotEmpty) {
      parameters = (parameters.isNotEmpty ? parameters + ',' : '') +
          '[${optPositionalParameters.join(',')}]';
    }
    if (namedParameters.isNotEmpty) {
      parameters = (parameters.isNotEmpty ? parameters + ',' : '') +
          '{${namedParameters.join(',')}}';
    }

    return parameters;
  }

  /// Generate Dart code for the OmnibusMessage.data construction.
  String dataMap() {
    var mapValues = [];

    // Read properties.
    properties.forEach((OmnibuildProperty property) {
      // Add parameter to map values.
      mapValues.add(
          "'${property.name}': ${property.enumeration ? '${property.name}.index' : property.name}");
    });

    return "{${mapValues.join(',')}}";
  }

  /// Commented documentation
  String get commentedDocumentation =>
      _prefixEachLine(documentation, '/// ', false);

  /// Get super class (OmnibusEvent, OmnibusRequest or OmnibusResponse).
  String get superClass => type == OmnibusMessageType.event
      ? 'OmnibusEvent'
      : (type == OmnibusMessageType.request
          ? 'OmnibusRequest'
          : 'OmnibusResponse');

  /// Get super class constructor (checked or unchecked)
  String get superConstructor =>
      supportsUncheckedInit() ? 'super.unchecked' : 'super';
}

/// Message specification information
class OmnibuildSpecFile {
  /// Major version
  final int version;

  /// Parent package
  final String package;

  /// Copyright
  String copyright;

  /// Parent library
  String library;

  /// Description
  String description;

  /// Messages
  List<OmnibuildMessage> messages = new List<OmnibuildMessage>();

  /// Construct from YAML data.
  OmnibuildSpecFile.fromYaml(String yamlString, this.version, this.package) {
    // Should pass by reference.
    _parseYaml(this, yamlString);
  }

  /// Generate code.
  String generateCode() {
    var code = _omnibuildTemplate.renderString(this);

    // Generate formatted code.
    //
    // Note: an exception will be raised if the user specifies illegal types or
    // default values in the specification. In this case the code will fail and
    // the exception will propagate to higher-level mechanisms.
    var formatter = new DartFormatter();

    // Return generated code.
    return formatter.format(code.toString());
  }

  /// Commented copyright
  String get commentedCopyright => _prefixEachLine(copyright, '// ', true);

  /// Commented description
  String get commentedDescription => _prefixEachLine(description, '/// ', true);
}
